# Copyright 2015 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from magnumclient.common import cliutils as utils
from magnumclient.common import utils as magnum_utils
from magnumclient import exceptions
from magnumclient.i18n import _

import os


DEPRECATION_MESSAGE = (
    'WARNING: Bay commands are deprecated and will be removed in a future '
    'release.\nUse cluster commands to avoid seeing this message.')


def _show_bay(bay):
    del bay._info['links']
    utils.print_dict(bay._info)


@utils.arg('--marker',
           metavar='<marker>',
           default=None,
           help=_('The last bay UUID of the previous page; '
                  'displays list of bays after "marker".'))
@utils.arg('--limit',
           metavar='<limit>',
           type=int,
           help=_('Maximum number of bays to return.'))
@utils.arg('--sort-key',
           metavar='<sort-key>',
           help=_('Column to sort results by.'))
@utils.arg('--sort-dir',
           metavar='<sort-dir>',
           choices=['desc', 'asc'],
           help=_('Direction to sort. "asc" or "desc".'))
@utils.arg('--fields',
           default=None,
           metavar='<fields>',
           help=_('Comma-separated list of fields to display. '
                  'Available fields: uuid, name, baymodel_id, stack_id, '
                  'status, master_count, node_count, links, bay_create_timeout'
                  )
           )
@utils.deprecated(DEPRECATION_MESSAGE)
def do_bay_list(cs, args):
    """Print a list of available bays.

    (Deprecated in favor of cluster-list.)
    """
    bays = cs.bays.list(marker=args.marker, limit=args.limit,
                        sort_key=args.sort_key,
                        sort_dir=args.sort_dir)
    columns = ['uuid', 'name', 'node_count', 'master_count', 'status']
    columns += utils._get_list_table_columns_and_formatters(
        args.fields, bays,
        exclude_fields=(c.lower() for c in columns))[0]
    utils.print_list(bays, columns,
                     {'versions': magnum_utils.print_list_field('versions')},
                     sortby_index=None)


@utils.deprecated(DEPRECATION_MESSAGE)
@utils.arg('--name',
           metavar='<name>',
           help=_('Name of the bay to create.'))
@utils.arg('--baymodel',
           required=True,
           metavar='<baymodel>',
           help=_('ID or name of the baymodel.'))
@utils.arg('--node-count',
           metavar='<node-count>',
           type=int,
           default=1,
           help=_('The bay node count.'))
@utils.arg('--master-count',
           metavar='<master-count>',
           type=int,
           default=1,
           help=_('The number of master nodes for the bay.'))
@utils.arg('--discovery-url',
           metavar='<discovery-url>',
           help=_('Specifies custom discovery url for node discovery.'))
@utils.arg('--timeout',
           metavar='<timeout>',
           type=int,
           default=60,
           help=_('The timeout for bay creation in minutes. The default '
                  'is 60 minutes.'))
def do_bay_create(cs, args):
    """Create a bay.

    (Deprecated in favor of cluster-create.)
    """
    baymodel = cs.baymodels.get(args.baymodel)

    opts = {}
    opts['name'] = args.name
    opts['baymodel_id'] = baymodel.uuid
    opts['node_count'] = args.node_count
    opts['master_count'] = args.master_count
    opts['discovery_url'] = args.discovery_url
    opts['bay_create_timeout'] = args.timeout
    try:
        bay = cs.bays.create(**opts)
        # support for non-async in 1.1
        if args.magnum_api_version and args.magnum_api_version == '1.1':
            _show_bay(bay)
        else:
            uuid = str(bay._info['uuid'])
            print("Request to create bay %s has been accepted." % uuid)
    except Exception as e:
        print("Create for bay %s failed: %s" %
              (opts['name'], e))


@utils.arg('bay',
           metavar='<bay>',
           nargs='+',
           help=_('ID or name of the (bay)s to delete.'))
@utils.deprecated(DEPRECATION_MESSAGE)
def do_bay_delete(cs, args):
    """Delete specified bay.

    (Deprecated in favor of cluster-delete.)
    """
    for id in args.bay:
        try:
            cs.bays.delete(id)
            print("Request to delete bay %s has been accepted." %
                  id)
        except Exception as e:
            print("Delete for bay %(bay)s failed: %(e)s" %
                  {'bay': id, 'e': e})


@utils.arg('bay',
           metavar='<bay>',
           help=_('ID or name of the bay to show.'))
@utils.arg('--long',
           action='store_true', default=False,
           help=_('Display extra associated Baymodel info.'))
@utils.deprecated(DEPRECATION_MESSAGE)
def do_bay_show(cs, args):
    """Show details about the given bay.

    (Deprecated in favor of cluster-show.)
    """
    bay = cs.bays.get(args.bay)
    if args.long:
        baymodel = cs.baymodels.get(bay.baymodel_id)
        del baymodel._info['links'], baymodel._info['uuid']

        for key in baymodel._info:
            if 'baymodel_' + key not in bay._info:
                bay._info['baymodel_' + key] = baymodel._info[key]
    _show_bay(bay)


@utils.arg('bay', metavar='<bay>', help=_("UUID or name of bay"))
@utils.arg('--rollback',
           action='store_true', default=False,
           help=_('Rollback bay on update failure.'))
@utils.arg(
    'op',
    metavar='<op>',
    choices=['add', 'replace', 'remove'],
    help=_("Operations: 'add', 'replace' or 'remove'"))
@utils.arg(
    'attributes',
    metavar='<path=value>',
    nargs='+',
    action='append',
    default=[],
    help=_("Attributes to add/replace or remove "
           "(only PATH is necessary on remove)"))
@utils.deprecated(DEPRECATION_MESSAGE)
def do_bay_update(cs, args):
    """Update information about the given bay.

    (Deprecated in favor of cluster-update.)
    """
    if args.rollback and args.magnum_api_version and \
            args.magnum_api_version in ('1.0', '1.1', '1.2'):
        raise exceptions.CommandError(
            "Rollback is not supported in API v%s. "
            "Please use API v1.3+." % args.magnum_api_version)
    patch = magnum_utils.args_array_to_patch(args.op, args.attributes[0])
    bay = cs.bays.update(args.bay, patch, args.rollback)
    if args.magnum_api_version and args.magnum_api_version == '1.1':
        _show_bay(bay)
    else:
        print("Request to update bay %s has been accepted." % args.bay)


@utils.arg('bay',
           metavar='<bay>',
           help=_('ID or name of the bay to retrieve config.'))
@utils.arg('--dir',
           metavar='<dir>',
           default='.',
           help=_('Directory to save the certificate and config files.'))
@utils.arg('--force',
           action='store_true', default=False,
           help=_('Overwrite files if existing.'))
@utils.deprecated(DEPRECATION_MESSAGE)
def do_bay_config(cs, args):
    """Configure native client to access bay.

    You can source the output of this command to get the native client of the
    corresponding COE configured to access the bay.

    Example: eval $(magnum bay-config <bay-name>).

    (Deprecated in favor of cluster-config.)
    """
    args.dir = os.path.abspath(args.dir)
    bay = cs.bays.get(args.bay)
    if (hasattr(bay, 'api_address') and bay.api_address is None):
        print("WARNING: The bay's api_address is not known yet.")
    baymodel = cs.baymodels.get(bay.baymodel_id)
    opts = {
        'cluster_uuid': bay.uuid,
    }

    if not baymodel.tls_disabled:
        tls = magnum_utils.generate_csr_and_key()
        tls['ca'] = cs.certificates.get(**opts).pem
        opts['csr'] = tls['csr']
        tls['cert'] = cs.certificates.create(**opts).pem
        for k in ('key', 'cert', 'ca'):
            fname = "%s/%s.pem" % (args.dir, k)
            if os.path.exists(fname) and not args.force:
                raise Exception("File %s exists, aborting." % fname)
            else:
                f = open(fname, "w")
                f.write(tls[k])
                f.close()

    print(magnum_utils.config_cluster(bay, baymodel, cfg_dir=args.dir,
                                      force=args.force))



# Copyright 2014 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from magnumclient.v1 import baseunit


CREATION_ATTRIBUTES = baseunit.CREATION_ATTRIBUTES
CREATION_ATTRIBUTES.append('baymodel_id')
CREATION_ATTRIBUTES.append('bay_create_timeout')


class Bay(baseunit.BaseTemplate):
    template_name = "Bays"


class BayManager(baseunit.BaseTemplateManager):
    resource_class = Bay
    template_name = 'bays'



# Copyright 2015 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import testtools
from testtools import matchers

from magnumclient import exceptions
from magnumclient.tests import utils
from magnumclient.v1 import bays


BAY1 = {'id': 123,
        'uuid': '66666666-7777-8888-9999-000000000001',
        'name': 'bay1',
        'baymodel_id': 'e74c40e0-d825-11e2-a28f-0800200c9a61',
        'stack_id': '5d12f6fd-a196-4bf0-ae4c-1f639a523a51',
        'api_address': '172.17.2.1',
        'node_addresses': ['172.17.2.3'],
        'node_count': 2,
        'master_count': 1,
        }
BAY2 = {'id': 124,
        'uuid': '66666666-7777-8888-9999-000000000002',
        'name': 'bay2',
        'baymodel_id': 'e74c40e0-d825-11e2-a28f-0800200c9a62',
        'stack_id': '5d12f6fd-a196-4bf0-ae4c-1f639a523a52',
        'api_address': '172.17.2.2',
        'node_addresses': ['172.17.2.4'],
        'node_count': 2,
        'master_count': 1,
        }

CREATE_BAY = copy.deepcopy(BAY1)
del CREATE_BAY['id']
del CREATE_BAY['uuid']
del CREATE_BAY['stack_id']
del CREATE_BAY['api_address']
del CREATE_BAY['node_addresses']

UPDATED_BAY = copy.deepcopy(BAY1)
NEW_NAME = 'newbay'
UPDATED_BAY['name'] = NEW_NAME

fake_responses = {
    '/v1/bays':
    {
        'GET': (
            {},
            {'bays': [BAY1, BAY2]},
        ),
        'POST': (
            {},
            CREATE_BAY,
        ),
    },
    '/v1/bays/%s' % BAY1['id']:
    {
        'GET': (
            {},
            BAY1
        ),
        'DELETE': (
            {},
            None,
        ),
        'PATCH': (
            {},
            UPDATED_BAY,
        ),
    },
    '/v1/bays/%s/?rollback=True' % BAY1['id']:
        {
            'PATCH': (
                {},
                UPDATED_BAY,
            ),
        },
    '/v1/bays/%s' % BAY1['name']:
    {
        'GET': (
            {},
            BAY1
        ),
        'DELETE': (
            {},
            None,
        ),
        'PATCH': (
            {},
            UPDATED_BAY,
        ),
    },
    '/v1/bays/?limit=2':
    {
        'GET': (
            {},
            {'bays': [BAY1, BAY2]},
        ),
    },
    '/v1/bays/?marker=%s' % BAY2['uuid']:
    {
        'GET': (
            {},
            {'bays': [BAY1, BAY2]},
        ),
    },
    '/v1/bays/?limit=2&marker=%s' % BAY2['uuid']:
    {
        'GET': (
            {},
            {'bays': [BAY1, BAY2]},
        ),
    },
    '/v1/bays/?sort_dir=asc':
    {
        'GET': (
            {},
            {'bays': [BAY1, BAY2]},
        ),
    },
    '/v1/bays/?sort_key=uuid':
    {
        'GET': (
            {},
            {'bays': [BAY1, BAY2]},
        ),
    },
    '/v1/bays/?sort_key=uuid&sort_dir=desc':
    {
        'GET': (
            {},
            {'bays': [BAY2, BAY1]},
        ),
    },
}


class BayManagerTest(testtools.TestCase):

    def setUp(self):
        super(BayManagerTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses)
        self.mgr = bays.BayManager(self.api)

    def test_bay_list(self):
        bays = self.mgr.list()
        expect = [
            ('GET', '/v1/bays', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(bays, matchers.HasLength(2))

    def _test_bay_list_with_filters(self, limit=None, marker=None,
                                    sort_key=None, sort_dir=None,
                                    detail=False, expect=[]):
        bays_filter = self.mgr.list(limit=limit, marker=marker,
                                    sort_key=sort_key,
                                    sort_dir=sort_dir,
                                    detail=detail)
        self.assertEqual(expect, self.api.calls)
        self.assertThat(bays_filter, matchers.HasLength(2))

    def test_bay_list_with_limit(self):
        expect = [
            ('GET', '/v1/bays/?limit=2', {}, None),
        ]
        self._test_bay_list_with_filters(
            limit=2,
            expect=expect)

    def test_bay_list_with_marker(self):
        expect = [
            ('GET', '/v1/bays/?marker=%s' % BAY2['uuid'], {}, None),
        ]
        self._test_bay_list_with_filters(
            marker=BAY2['uuid'],
            expect=expect)

    def test_bay_list_with_marker_limit(self):
        expect = [
            ('GET', '/v1/bays/?limit=2&marker=%s' % BAY2['uuid'], {}, None),
        ]
        self._test_bay_list_with_filters(
            limit=2, marker=BAY2['uuid'],
            expect=expect)

    def test_bay_list_with_sort_dir(self):
        expect = [
            ('GET', '/v1/bays/?sort_dir=asc', {}, None),
        ]
        self._test_bay_list_with_filters(
            sort_dir='asc',
            expect=expect)

    def test_bay_list_with_sort_key(self):
        expect = [
            ('GET', '/v1/bays/?sort_key=uuid', {}, None),
        ]
        self._test_bay_list_with_filters(
            sort_key='uuid',
            expect=expect)

    def test_bay_list_with_sort_key_dir(self):
        expect = [
            ('GET', '/v1/bays/?sort_key=uuid&sort_dir=desc', {}, None),
        ]
        self._test_bay_list_with_filters(
            sort_key='uuid', sort_dir='desc',
            expect=expect)

    def test_bay_show_by_id(self):
        bay = self.mgr.get(BAY1['id'])
        expect = [
            ('GET', '/v1/bays/%s' % BAY1['id'], {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(BAY1['name'], bay.name)
        self.assertEqual(BAY1['baymodel_id'], bay.baymodel_id)

    def test_bay_show_by_name(self):
        bay = self.mgr.get(BAY1['name'])
        expect = [
            ('GET', '/v1/bays/%s' % BAY1['name'], {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(BAY1['name'], bay.name)
        self.assertEqual(BAY1['baymodel_id'], bay.baymodel_id)

    def test_bay_create(self):
        bay = self.mgr.create(**CREATE_BAY)
        expect = [
            ('POST', '/v1/bays', {}, CREATE_BAY),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(bay)

    def test_bay_create_with_discovery_url(self):
        bay_with_discovery = dict()
        bay_with_discovery.update(CREATE_BAY)
        bay_with_discovery['discovery_url'] = 'discovery_url'
        bay = self.mgr.create(**bay_with_discovery)
        expect = [
            ('POST', '/v1/bays', {}, bay_with_discovery),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(bay)

    def test_bay_create_with_bay_create_timeout(self):
        bay_with_timeout = dict()
        bay_with_timeout.update(CREATE_BAY)
        bay_with_timeout['bay_create_timeout'] = '15'
        bay = self.mgr.create(**bay_with_timeout)
        expect = [
            ('POST', '/v1/bays', {}, bay_with_timeout),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(bay)

    def test_bay_create_fail(self):
        CREATE_BAY_FAIL = copy.deepcopy(CREATE_BAY)
        CREATE_BAY_FAIL["wrong_key"] = "wrong"
        self.assertRaisesRegex(exceptions.InvalidAttribute,
                               ("Key must be in %s" %
                                ','.join(bays.CREATION_ATTRIBUTES)),
                               self.mgr.create, **CREATE_BAY_FAIL)
        self.assertEqual([], self.api.calls)

    def test_bay_delete_by_id(self):
        bay = self.mgr.delete(BAY1['id'])
        expect = [
            ('DELETE', '/v1/bays/%s' % BAY1['id'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertIsNone(bay)

    def test_bay_delete_by_name(self):
        bay = self.mgr.delete(BAY1['name'])
        expect = [
            ('DELETE', '/v1/bays/%s' % BAY1['name'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertIsNone(bay)

    def test_bay_update(self):
        patch = {'op': 'replace',
                 'value': NEW_NAME,
                 'path': '/name'}
        bay = self.mgr.update(id=BAY1['id'], patch=patch)
        expect = [
            ('PATCH', '/v1/bays/%s' % BAY1['id'], {}, patch),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(NEW_NAME, bay.name)

    def test_bay_update_with_rollback(self):
        patch = {'op': 'replace',
                 'value': NEW_NAME,
                 'path': '/name'}
        bay = self.mgr.update(id=BAY1['id'], patch=patch, rollback=True)
        expect = [
            ('PATCH', '/v1/bays/%s/?rollback=True' % BAY1['id'], {}, patch),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(NEW_NAME, bay.name)
